-- 
-- Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
--
-- 
-- Filename: diy_calculator.vhd
--
-- Function: the DIY calculator overall design
-- 
--
-- Changelog
--
--  0.0   25.Jul.2006   JH   new
--  0.1   31.Aug.2006   JH   integrate LCD controller
--  0.2    5.Sep.2006   JH   integrate UART
--  0.3    7.Sep.2006   JH   add monitor RAM
--  0.4    8.Sep.2006   JH   add HW revision port
--  0.5   13.Sep.2006   JH   add hard/soft reset flip flop in IRQ SRC port
--                           error in cpu_irq solved
--  0.6   15.Sep.2006   JH   add IRQ controller
--  0.7   18.Sep.2006   JH   add CPU clock control
--  0.8   25.Sep.2006   JH   add SPI controller
--         3.Oct.2006   JH   add reset synchronizer
--
-- Please set the constant "REVISION" so that it reflects the version!
--
--


library ieee;
use ieee.std_logic_1164.all;

entity diy_calculator is 
  port (
    -- global signals
    nSysRst  : in std_logic;
    CLK25    : in std_logic;
    -- CPU signals
    Addr     : out std_logic_vector (15 downto 0);
    Data     : inout std_logic_vector (7 downto 0);
    nRD      : out std_logic;
    nWR      : out std_logic;
    nRAMext  : out std_logic;
    nROMext  : out std_logic;
    nIO      : out std_logic;
    nIRQ     : in std_logic;
    nIACK    : out std_logic;
    nWAIT    : out std_logic;
    M1       : out std_logic;
    JP_ROM   : in std_logic;
    JP_RAM   : in std_logic;
    -- peripherals
    LED      : out std_logic_vector (7 downto 0);
    Beep     : out std_logic;
    LCDdata  : out std_logic_vector (7 downto 0);
    LCDen    : out std_logic;
    LCDrs    : out std_logic;
    LCDrnw   : out std_logic;
    -- keypad matrix
--    KeyXin    : in std_logic_vector (8 downto 0);
--    KeyYin    : in std_logic_vector (8 downto 0);
--    KeyClkOut : out std_logic;
    KeyYout  : out std_logic_vector (2 downto 0);
    KeyXin   : in std_logic_vector (3 downto 0);
    -- UART
    UartTxD  : out std_logic;
    UartRxD  : in std_logic;
    UartRTS  : out std_logic;
    UartCTS  : in std_logic;
    -- SPI to serial flash
    SPIclk   : out std_logic;
    SPIcs    : out std_logic;
    SPIdin   : in std_logic;
    SPIdout  : out std_logic
  );
end diy_calculator;

architecture structure of diy_calculator is
  constant REVISION : std_logic_vector (7 downto 0) := x"08";
  component cpu
    port (
      CLK    : in std_logic;
      nRST   : in std_logic;
      d_in   : in std_logic_vector (7 downto 0);
      d_out  : out std_logic_vector (7 downto 0);
      addr   : out std_logic_vector (15 downto 0);
      nIRQ   : in std_logic;
      nIACK  : out std_logic;
      nWAIT  : in std_logic;
      M1     : out std_logic;
      nRD    : out std_logic;
      nWR    : out std_logic;
      ddrive : out std_logic
    );
  end component;

  component clock_generator
    port (
      rst      : in std_logic;
      clk_in   : in std_logic;
      sel      : in std_logic;
      nWR      : in std_logic;
      ddrive   : in std_logic;
      d_in     : in std_logic_vector (7 downto 0);
      d_out    : out std_logic_vector (7 downto 0);
      cpu_clk  : out std_logic;
      uart_clk : out std_logic;
      lcd_clk  : out std_logic;
      key_clk  : out std_logic;
      spi_clk  : out std_logic
    );
  end component;

  component address_decoder
    port (
      addr  : in std_logic_vector (15 downto 0);
      rom   : out std_logic;
      ram   : out std_logic;
      io    : out std_logic;
      key   : out std_logic;
      ver   : out std_logic;
      irq   : out std_logic;
      lcd   : out std_logic;
      led   : out std_logic;
      uart  : out std_logic;
      spi   : out std_logic;
      cpu_clk : out std_logic;
      mram  : out std_logic
    );
  end component;

  component rom
    port (
      Address    : in std_logic_vector (12 downto 0);
      OutClock   : in std_logic;
      OutClockEn : in std_logic;
      Reset      : in std_logic;
      Q          : out std_logic_vector (7 downto 0)
    );
  end component;

  component ram
    port (
      Clock   : in std_logic;
      ClockEn : in std_logic;
      Reset   : in std_logic;
      WE      : in std_logic;
      Address : in std_logic_vector (10 downto 0);
      Data    : in std_logic_vector (7 downto 0);
      Q       : out std_logic_vector (7 downto 0)
    );
  end component;

  component mram
    port (
      Clock   : in std_logic;
      ClockEn : in std_logic;
      Reset   : in std_logic;
      WE      : in std_logic;
      Address : in std_logic_vector (8 downto 0);
      Data    : in std_logic_vector (7 downto 0);
      Q       : out std_logic_vector (7 downto 0)
    );
  end component;

  component uart
    port (
      clk     : in std_logic;  -- 3 times the baud rate
      rst     : in std_logic;
      TxD     : out std_logic;
      RxD     : in std_logic;
      nRTS    : out std_logic;
      nCTS    : in std_logic;
      sel     : in std_logic;
      nWR     : in std_logic;
      nRD     : in std_logic;
      RS      : in std_logic;  -- 0 .. data, 1 .. ctrl/status
      d_in    : in std_logic_vector(7 downto 0);
      d_out   : out std_logic_vector(7 downto 0);
      irq     : out std_logic
    );
  end component;

  component lcd_ctrl
    port (
      rst   : in std_logic;
      clk   : in std_logic;
      sel   : in std_logic;
      nWR   : in std_logic;
      d_in  : in std_logic_vector (7 downto 0);
      busy  : out std_logic;
      en    : out std_logic;
      rs    : out std_logic;
      rnw   : out std_logic;
      beep  : out std_logic;
      d_out : out std_logic_vector (7 downto 0)
    );
  end component;

  component keypad_ctrl
    port (
      clk     : in std_logic;
      rst     : in std_logic;
      sel     : in std_logic;
--      clk_out : out std_logic;
--      X_in    : in std_logic_vector (8 downto 0);
--      Y_in    : in std_logic_vector (8 downto 0);
      Y_out   : out std_logic_vector (2 downto 0);
      X_in    : in std_logic_vector (3 downto 0);
      beep    : out std_logic;
      run     : out std_logic;
      step    : out std_logic;
      d_out   : out std_logic_vector (7 downto 0)
    );
  end component;

  component irq_ctrl
    port (
      clk      : in std_logic;
      rst      : in std_logic;
      sel      : in std_logic;
      irq_in   : in std_logic;
      uart     : in std_logic;
      key_run  : in std_logic;
      key_step : in std_logic;
      irq_out  : out std_logic;
      d_out    : out std_logic_vector (7 downto 0)
    );
  end component;

  component spi_ctrl
    port (
      clk      : in std_logic;
      rst      : in std_logic;
      spi_clk  : out std_logic;
      spi_cs   : out std_logic;
      spi_din  : in std_logic;
      spi_dout : out std_logic;
      sel      : in std_logic;
      nWR      : in std_logic;
      nRD      : in std_logic;
      addr     : in std_logic_vector (1 downto 0);
      d_in     : in std_logic_vector (7 downto 0);
      d_out    : out std_logic_vector (7 downto 0);
      test     : out std_logic
    );
  end component;

  component dff8
    port (
      rst   : in std_logic;
      clk   : in std_logic;
      d_in  : in std_logic_vector (7 downto 0);
      d_out : out std_logic_vector (7 downto 0)
    );
  end component;

  signal sysrst, rst1, nwait_int : std_logic;
  signal nrd_int, nwr_int, wr_int, ddrive : std_logic;

  signal cpu_clk, cpu_clk_in, cpu_irq : std_logic;
  signal cpu_d_in, cpu_d_out : std_logic_vector (7 downto 0);
  signal addr_int : std_logic_vector (15 downto 0);

  signal rom_sel, rom_int : std_logic;
  signal rom_data : std_logic_vector (7 downto 0);

  signal ram_sel, ram_int, wr_ram : std_logic;
  signal ram_data : std_logic_vector (7 downto 0);

  signal lcd_clk, lcd_sel, lcd_beep, lcd_busy, lcd_en : std_logic;
  signal lcd_d_out : std_logic_vector (7 downto 0);

  signal key_clk, key_sel, key_beep, key_run, key_step : std_logic;
  signal key_data : std_logic_vector (7 downto 0);

  signal uart_clk, uart_sel, uart_irq : std_logic;
  signal uart_data : std_logic_vector (7 downto 0);
  signal uart_rxd, uart_txd, uart_rts, uart_cts : std_logic;

  signal led_clk, led_sel : std_logic;
  signal led_data : std_logic_vector (7 downto 0);
  alias beep_int : std_logic is led_data(7);

  signal irq_ctrl_out, irq_sel, irq_in : std_logic;
  signal irq_data : std_logic_vector (7 downto 0);

  signal mram_sel, wr_mram : std_logic;
  signal mram_data : std_logic_vector (7 downto 0);

  signal clk_sel : std_logic;
  signal clk_data : std_logic_vector (7 downto 0);

  signal spi_clk, spi_sel, spi_clk_int, spi_cs_int, spi_test : std_logic;
  signal spi_data : std_logic_vector (7 downto 0);

  signal io_sel : std_logic;

  signal ver_sel : std_logic;

begin
  Addr <= addr_int;
  nRD <= nrd_int;
  nWR <= nwr_int;
  wr_int <= not nwr_int;
  wr_ram <= wr_int and ram_sel;
  wr_mram <= wr_int and mram_sel;
  irq_in <= not nIRQ;
  UartTxD <= uart_txd;
  uart_rxd <= UartRxD;
  UartRTS <= uart_rts;
  uart_cts <= UartCTS;
  SPIclk <= spi_clk_int;
  SPIcs <= not spi_cs_int;

  -- synchronize reset
  process (nSysRst, CLK25)
  begin
    if nSysRst = '0' then
      rst1 <= '1';
      sysrst <= '1';
    elsif rising_edge (CLK25) then
      rst1 <= '0';
      sysrst <= rst1;
    end if;
  end process;

  -- external data bus
  Data(0) <= (cpu_d_out(0) and     ddrive)
          or (cpu_d_in(0)  and not(ddrive));
  Data(1) <= (cpu_d_out(1) and     ddrive)
          or (cpu_d_in(1)  and not(ddrive));
  Data(2) <= (cpu_d_out(2) and     ddrive)
          or (cpu_d_in(2)  and not(ddrive));
  Data(3) <= (cpu_d_out(3) and     ddrive)
          or (cpu_d_in(3)  and not(ddrive));
  Data(4) <= (cpu_d_out(4) and     ddrive)
          or (cpu_d_in(4)  and not(ddrive));
  Data(5) <= (cpu_d_out(5) and     ddrive)
          or (cpu_d_in(5)  and not(ddrive));
  Data(6) <= (cpu_d_out(6) and     ddrive)
          or (cpu_d_in(6)  and not(ddrive));
  Data(7) <= (cpu_d_out(7) and     ddrive)
          or (cpu_d_in(7)  and not(ddrive));

  -- CPU clock
  cpu_clk_in <= cpu_clk;

  -- select int/ext ROM
  rom_int <= rom_sel and JP_ROM;
  nROMext <= not (rom_sel and not JP_ROM);

  -- select int/ext RAM
  ram_int <= ram_sel and JP_RAM;
  nRAMext <= not (ram_sel and not JP_RAM);

  -- invert external IO select
--  nIO <= not io_sel;
--  nIO <= uart_rxd;
  nIO <= spi_test;

  -- beeper output
  Beep <= beep_int or lcd_beep or key_beep;

  -- LED port is clocked with rising edge
  led_clk <= not (led_sel) or nwr_int;
  LED <= not (led_data);

  -- LCD
  LCDen <= lcd_en;
  LCDdata <= lcd_d_out;

  -- WAIT generation
  nwait_int <= not (lcd_busy and lcd_sel);
--  nWAIT <= nwait_int;
--  nWAIT <= not mram_sel;
  nWAIT <= uart_txd;

  -- IRQ source
  cpu_irq <= not irq_ctrl_out;

  -- internal data bus switching
  cpu_d_in(0) <= (rom_data(0)  and rom_int)
              or (ram_data(0)  and ram_int)
              or (key_data(0)  and key_sel)
              or (REVISION(0)  and ver_sel)
              or (irq_data(0)  and irq_sel)
              or (clk_data(0)  and clk_sel)
              or (uart_data(0) and uart_sel)
              or (spi_data(0)  and spi_sel)
              or (mram_data(0) and mram_sel);
--              or (Data(0)      and not nROMext)
--              or (Data(0)      and not nRAMext);

  cpu_d_in(1) <= (rom_data(1)  and rom_int)
              or (ram_data(1)  and ram_int)
              or (key_data(1)  and key_sel)
              or (REVISION(1)  and ver_sel)
              or (irq_data(1)  and irq_sel)
              or (clk_data(1)  and clk_sel)
              or (uart_data(1) and uart_sel)
              or (spi_data(1)  and spi_sel)
              or (mram_data(1) and mram_sel);
--              or (Data(1)      and not nROMext)
--              or (Data(1)      and not nRAMext);

  cpu_d_in(2) <= (rom_data(2)  and rom_int)
              or (ram_data(2)  and ram_int)
              or (key_data(2)  and key_sel)
              or (REVISION(2)  and ver_sel)
              or (irq_data(2)  and irq_sel)
              or (clk_data(2)  and clk_sel)
              or (uart_data(2) and uart_sel)
              or (spi_data(2)  and spi_sel)
              or (mram_data(2) and mram_sel);
--              or (Data(2)      and not nROMext)
--              or (Data(2)      and not nRAMext);

  cpu_d_in(3) <= (rom_data(3)  and rom_int)
              or (ram_data(3)  and ram_int)
              or (key_data(3)  and key_sel)
              or (REVISION(3)  and ver_sel)
              or (irq_data(3)  and irq_sel)
              or (clk_data(3)  and clk_sel)
              or (uart_data(3) and uart_sel)
              or (spi_data(3)  and spi_sel)
              or (mram_data(3) and mram_sel);
--              or (Data(3)      and not nROMext)
--              or (Data(3)      and not nRAMext);

  cpu_d_in(4) <= (rom_data(4)  and rom_int)
              or (ram_data(4)  and ram_int)
              or (key_data(4)  and key_sel)
              or (REVISION(4)  and ver_sel)
              or (irq_data(4)  and irq_sel)
              or (clk_data(4)  and clk_sel)
              or (uart_data(4) and uart_sel)
              or (spi_data(4)  and spi_sel)
              or (mram_data(4) and mram_sel);
--              or (Data(4)      and not nROMext)
--              or (Data(4)      and not nRAMext);

  cpu_d_in(5) <= (rom_data(5)  and rom_int)
              or (ram_data(5)  and ram_int)
              or (key_data(5)  and key_sel)
              or (REVISION(5)  and ver_sel)
              or (irq_data(5)  and irq_sel)
              or (clk_data(5)  and clk_sel)
              or (uart_data(5) and uart_sel)
              or (spi_data(5)  and spi_sel)
              or (mram_data(5) and mram_sel);
--              or (Data(5)      and not nROMext)
--              or (Data(5)      and not nRAMext);

  cpu_d_in(6) <= (rom_data(6)  and rom_int)
              or (ram_data(6)  and ram_int)
              or (key_data(6)  and key_sel)
              or (REVISION(6)  and ver_sel)
              or (irq_data(6)  and irq_sel)
              or (clk_data(6)  and clk_sel)
              or (uart_data(6) and uart_sel)
              or (spi_data(6)  and spi_sel)
              or (mram_data(6) and mram_sel);
--              or (Data(6)      and not nROMext)
--              or (Data(6)      and not nRAMext);

  cpu_d_in(7) <= (rom_data(7)  and rom_int)
              or (ram_data(7)  and ram_int)
              or (key_data(7)  and key_sel)
              or (REVISION(7)  and ver_sel)
              or (irq_data(7)  and irq_sel)
              or (clk_data(7)  and clk_sel)
              or (uart_data(7) and uart_sel)
              or (spi_data(7)  and spi_sel)
              or (mram_data(7) and mram_sel);
--              or (Data(7)      and not nROMext)
--              or (Data(7)      and not nRAMext);

  my_cpu : cpu port map (
    CLK => cpu_clk_in,
    nRST => nSysRst,
    d_in => cpu_d_in,
    d_out => cpu_d_out,
    addr => addr_int,
    nIRQ => cpu_irq,
    nIACK => nIACK,
    nWAIT => nwait_int,
    M1 => M1,
    nRD => nrd_int,
    nWR => nwr_int,
    ddrive => ddrive
  );

  my_clock: clock_generator port map (
    rst => sysrst,
    clk_in => CLK25,
    sel => clk_sel,
    nWR => nwr_int,
    ddrive => ddrive,
    d_in => Data,
    d_out => clk_data,
    cpu_clk => cpu_clk,
    uart_clk => uart_clk,
    lcd_clk => lcd_clk,
    spi_clk => spi_clk,
    key_clk => key_clk
  );

  my_address: address_decoder port map (
    addr => addr_int,
    rom  => rom_sel,
    ram  => ram_sel,
    key  => key_sel,
    ver  => ver_sel,
    irq  => irq_sel,
    lcd  => lcd_sel,
    led  => led_sel,
    spi  => spi_sel,
    io   => io_sel,
    uart => uart_sel,
    cpu_clk => clk_sel,
    mram => mram_sel
  );

  my_rom: rom port map (
    Address => addr_int (12 downto 0),
    OutClock => CLK25,
    OutClockEn => '1',
    Reset => sysrst,
    Q => rom_data
  );

  my_ram: ram port map (
    Clock => CLK25,
    ClockEn => '1',
    Reset => sysrst,
    WE => wr_ram,
    Address => addr_int (10 downto 0),
    Data => Data,
    Q => ram_data
  );

  monitor_ram: mram port map (
    Clock => CLK25,
    ClockEn => '1',
    Reset => sysrst,
    WE => wr_mram,
    Address => addr_int (8 downto 0),
    Data => Data,
    Q => mram_data
  );

  my_uart: uart port map (
    clk => uart_clk,
    rst => sysrst,
    TxD => uart_txd,
    RxD => uart_rxd,
    nRTS => uart_rts,
    nCTS => uart_cts,
    sel => uart_sel,
    nWR => nwr_int,
    nRD => nrd_int,
    RS => addr_int(0),
    d_in => Data,
    d_out => uart_data,
    irq => uart_irq
  );

  my_lcd: lcd_ctrl port map (
    rst => sysrst,
    clk => lcd_clk,
    sel => lcd_sel,
    nWR => nwr_int,
    d_in => Data,
    busy => lcd_busy,
    en => lcd_en,
    rs => LCDrs,
    rnw => LCDrnw,
    beep => lcd_beep,
    d_out => lcd_d_out
  );

  my_keypad: keypad_ctrl port map (
    clk => key_clk,
    rst => sysrst,
    sel => key_sel,
--    clk_out => KeyClkOut,
--    X_in => KeyXin,
--    Y_in => KeyYin,
    Y_out => KeyYout,
    X_in => KeyXin,
    beep => key_beep,
    run => key_run,
    step => key_step,
    d_out => key_data
  );

  my_irq : irq_ctrl port map (
    clk => CLK25,
    rst => sysrst,
    sel => irq_sel,
    irq_in => irq_in,
    uart => uart_irq,
    key_run => key_run,
    key_step => key_step,
    irq_out => irq_ctrl_out,
    d_out => irq_data
  );

  my_spi: spi_ctrl port map (
    clk => spi_clk,
    rst => sysrst,
    spi_clk => spi_clk_int,
    spi_cs => spi_cs_int,
    spi_din => SPIdin,
    spi_dout => SPIdout,
    sel => spi_sel,
    nWR => nwr_int,
    nRD => nrd_int,
    addr => addr_int(1 downto 0),
    d_in => Data,
    d_out => spi_data,
    test => spi_test
  );

  my_led: dff8 port map (
    rst => sysrst,
    clk => led_clk,
    d_in => cpu_d_out,
    d_out => led_data
  );
end structure;
